
/* Copyright (c) 1996 Dennis P. Glatting
        dennis.glatting@plaintalk.bellevue.wa.us
    All rights reserved.

    The author disclaims all responsibility or liability with 
    respect to it's usage or its effect upon hardware or 
    computer systems, and maintains copyright as set out in
    the "Artistic" license document which accompanies 
    distributions of PERL 5. 

    This program generates Crack word character
    substitution rules to guess It also generates rule
    variations to pluralize words, reflect them,
    capitalize, and lower and upper case them. 

    In my experience, Crack commonly achieves a 20% guess
    rate. One of the algorithms users have adopted to thwart
    Crack is word character substitutions, for example,
    they may substitute a "*" or "0" for an "o" in the word
    "password". The dictionary rules distributed with
    Crack perform a limited set of substitutions to attack
    these algorithms. This program generates rules that
    takes that character substitution to extreme. 

    The rules generated by this program are not meant to
    augment the existing substitution rules distributed
    with the Crack (i.e., the rules created by
    mycroft@edu.mit.ai.gnu), rather they replace them,
    though I suggest putting them at the end of the rule file 
    due to the more productive nature of the existing rules. 

    Warning: This program generates over 900k rules,
    consuming over 60m of disk space. On my Sun IPC it takes 216
    seconds to apply a rule yielding no words against a 15mb
    (compressed) dictionary. At that rate, it would take a
    minimum of 369 years to perform a single Crack run. You may
    want to tweak the number of rules generated by this code to
    reduce that some. :) 

    Often, rules are rejected. A rule that specifies a lot of
    substitutions requires a word that has the necessary
    characters to substitute. The more substitutions, the
    more characters required, the less likely there will be a
    word in the dictionary containing all of the characters.
    An example of a rule that may be rejected is one that lower
    cases a word then specifies substitution for the
    characters "c" and "C" -- there is no "C". 
    
    Hint: Sort the rules by length from shortest to longest.
    
    There are other substations I'd like to try such as
    substituting " " for "-", "!", and others; however, the
    output file is already bit huge. :) 
    
 */


#include    <assert.h>


/* A silly little marco that computes 
    the number of elements in an array. */

#define SZ(XYZZY)   ( sizeof(XYZZY)/sizeof(XYZZY[0]))

/* This is a list of substitution patterns. 
    The first character is the character to 
    be replaced, subsequent characters are 
    its replacements. */

const char *    patterns[] = {  
    "s25$", "S25$", "a4", "A4", "e3", "E3", 
    "h4", "H4", "l|!/1", "L|!/1", 
    "i|!/1", "I|!/1", "o*0", "O*0", 
    "c(", "C(", 0
};

/* This is a list of case rules.
    For every substitution rule, a case rule
    is applied. The first case rule does not
    change the word. The second rule converts
    the word to lower case. The third rule
    converts the word to upper case. Finally, 
    the last case rule is an attempt to
    capitalize it. */
    
const char *    caseRule[] = {  "", "l", "u", "c", 0 };

/* This is a list of -- what I call --- reflection 
    rules. The first reflection rule does not 
    change the word. The second rule attempts to 
    pluralize the word. Finally, the last rule 
    "reflects" the word. */
    
const char *    reflectRule[] = { "", "p", "r", 0 };


/* This recursive function builds the substitution 
    components of rules. */

void
catit( char * buf, int idx ) {

    char    qBuf[ 1024 ];
    
    if( patterns[ idx ]) {
        
        int j, k;
        
        for( j = 1; j < patterns[idx][j]; ++j ) {
        
            sprintf( qBuf, "%s/%cs%c%c", buf, 
                    patterns[idx][0], patterns[idx][0], patterns[idx][j]);
            assert( strlen( qBuf ) < sizeof( qBuf ));
            catit( qBuf, idx + 1 );

            printf( "%s\n", qBuf );
        }
    }
}
        

int
main( void ) {

    char    tBuf[ 1024 ];
    int     i, j, k;

    for( i = 0; i < SZ(patterns); ++i ) {
    
        for( j = 0; reflectRule[ j ]; ++j ) {
        
            for( k = 0; caseRule[ k ]; ++k ) {
            
                sprintf( tBuf, ">2%s%s", caseRule[ k ], reflectRule[ j ]);
                catit( tBuf, i );
            }
        }
    }

    return 0;
}


